<sly-card
  :title="data.{{{widget.widget_id}}}.title"
  :subtitle="data.{{{widget.widget_id}}}.description"
  :options="data.{{{widget.widget_id}}}.options"
  :disabled="state.{{{widget.widget_id}}}.disabled"
  :collapsed="state.{{{widget.widget_id}}}.collapsed"
>
  <div>{{{widget._content}}}</div>
</sly-card>
